import numpy as np
import pandas as pd
import csv
import recommender_ibmws.recommender_functions as rf
import sys as sys
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import linear_kernel
from sklearn.metrics.pairwise import cosine_similarity


class Recommender():
    '''
    This Recommender uses hybrid approach ofContent-Based Filtering and Collaborative Filtering to 
    make recommendations.  
    The content base recommendation system developed relies on a user profile, text vectorization, 
    similarity calculation, ranking and recommendation, as well as handling of new users. 
    '''

    def __init__(self):
        
        self.interactions = None
        self.content = None
        
    def load_data(self,inter_path, content_path):
        """
        Loads interaction and content data from CSV files, maps email addresses to unique user IDs, 
        and returns the processed interaction and content dataframes.

        Parameters:
            inter_path (str): The filepath to the interactions CSV file.
            content_path (str): The filepath to the content CSV file.

        Returns:
            tuple: A tuple containing the processed interaction and content dataframes.
        """
        
        self.interactions, self.content = rf.load_data(inter_path, content_path)
        return self.interactions, self.content

    def make_content_recs(self, user_id, m=10):
        """
        Makes content-based recommendations for a given user.

        Parameters:
            user_id (int): The ID of the user for whom to make recommendations.
            m (int): The number of recommendations to make.

        Returns:
            tuple: A tuple containing the hybrid recommendations and their names.
            This function stores the following as attributes:
            n_users - the number of users (int)
            n_articles - the number of articles (int)
            num_interactions - the number of interactions (int)
            user_item - the user-item matrix (pandas dataframe)
            user_id - the user ID (int)
            m - the number of recommendations (int)
            user_profile - the user profile (str)
            interactions - the interaction dataframe (pandas dataframe)
            content - the content dataframe (pandas dataframe)
            hybrid_recs - the recommendations (list)
            hybrid_rec_names - the recommended article names (list)
        """
        # Create user-item matrix
        self.user_item = rf.create_user_item_matrix(self.interactions)
        
        # Store more inputs
        self.user_id = user_id
        self.m = m
        
        # Set up useful values to be used through the rest of the function
        self.n_users = self.user_item.shape[0]
        self.n_articles = self.user_item.shape[1]
        self.num_interactions = np.count_nonzero(~np.isnan(self.user_item))
        
        # Create user-item matrix only once
        user_item = rf.create_user_item_matrix(self.interactions)
        
        # Check if user has any interactions
        if user_id in user_item.index:
            user_articles = rf.get_user_articles(user_id, self.interactions)[0]
            
            if user_articles:
                # Build user profile based on the content of articles the user has interacted with
                user_profile = ' '.join(self.content[self.content['article_id'].isin(user_articles)]['doc_full_name'])
                
                tfidf = TfidfVectorizer(stop_words='english')
                tfidf_matrix = tfidf.fit_transform(self.content['doc_full_name'])
                user_profile_vector = tfidf.transform([user_profile])
                
                # Calculate content-based recommendations
                cosine_sim_content = cosine_similarity(user_profile_vector, tfidf_matrix)
                content_based_rec_indices = cosine_sim_content.argsort()[0][::-1]
                
                # Filter out articles the user has already interacted with
                content_based_recs = [self.content.iloc[i]['article_id'] for i in content_based_rec_indices 
                                    if self.content.iloc[i]['article_id'] not in user_articles][:m]
            else:
                content_based_recs = rf.get_top_article_ids(m, self.interactions)
        else:
            content_based_recs = rf.get_top_article_ids(m, self.interactions)

        # Collaborative Filtering
        if user_id in user_item.index:
            user_interactions = user_item.loc[user_id]
            similar_users = user_item.drop(user_id).apply(
                lambda row: cosine_similarity([user_interactions], [row])[0][0], axis=1)
            similar_users = similar_users.sort_values(ascending=False)
            
            collaborative_recs = []
            for sim_user in similar_users.index:
                sim_user_articles = user_item.loc[sim_user]
                new_recs = sim_user_articles[sim_user_articles == 1].index.difference(user_interactions[user_interactions == 1].index)
                collaborative_recs.extend(new_recs)
                if len(collaborative_recs) >= m:
                    break
            collaborative_recs = collaborative_recs[:m]
        else:
            collaborative_recs = rf.get_top_article_ids(m, self.interactions)

        # Hybrid Recommendations
        hybrid_recs = list(set(content_based_recs) | set(collaborative_recs))[:m]
        hybrid_rec_names = rf.get_article_names(hybrid_recs, self.interactions)
        
        return hybrid_recs, hybrid_rec_names

if __name__ == '__main__':
    import recommender as r

    #instantiate recommender
    rec = r.Recommender()
    
    # load data
    rec.load_data(inter_path='data/user-item-interactions.csv', content_path= 'data/articles_community.csv')

    # make recommendations
    rec.make_recommendations(8, m=10) # user in the dataset
    rec.make_recommendations(5150, m=10)# user not in dataset
    rec.n_users
    rec.n_articles


import unittest
from unittest.mock import patch, MagicMock
import pandas as pd
import numpy as np
from recommender_ibmws.recommender import Recommender  # Adjusted import for the Recommender class

class TestRecommender(unittest.TestCase):

    def setUp(self):
        # This method will run before each test. It's useful for setting up any shared state.
        self.recommender = Recommender()
        self.interactions_mock = pd.DataFrame({
            'user_id': [1, 2, 3, 4],
            'article_id': [101, 102, 103, 104],
            'interaction': [1, 1, 1, 1]
        })
        self.content_mock = pd.DataFrame({
            'article_id': [101, 102, 103, 104, 105],
            'doc_full_name': ['Article A', 'Article B', 'Article C', 'Article D', 'Article E']
        })

    @patch('recommender_ibmws.recommender.rf.load_data')
    def test_load_data(self, mock_load_data):
        # Mock the rf.load_data function
        mock_load_data.return_value = (self.interactions_mock, self.content_mock)
        
        interactions, content = self.recommender.load_data('interactions.csv', 'content.csv')
        
        mock_load_data.assert_called_once_with('interactions.csv', 'content.csv')
        pd.testing.assert_frame_equal(interactions, self.interactions_mock)
        pd.testing.assert_frame_equal(content, self.content_mock)

    @patch('recommender_ibmws.recommender.rf.create_user_item_matrix')
    @patch('recommender_ibmws.recommender.rf.get_user_articles')
    @patch('recommender_ibmws.recommender.rf.get_top_article_ids')
    @patch('recommender_ibmws.recommender.rf.get_article_names')
    def test_make_content_recs(self, mock_get_article_names, mock_get_top_article_ids, mock_get_user_articles, mock_create_user_item_matrix):
        # Mock functions
        mock_create_user_item_matrix.return_value = pd.DataFrame({
            101: [1, 0, 0, 0],
            102: [0, 1, 0, 0],
            103: [0, 0, 1, 0],
            104: [0, 0, 0, 1],
        }, index=[1, 2, 3, 4])
        
        mock_get_user_articles.return_value = ([101], ['Article A'])
        mock_get_top_article_ids.return_value = [105, 103, 104]
        mock_get_article_names.return_value = ['Article E', 'Article C', 'Article D']
        
        self.recommender.interactions = self.interactions_mock
        self.recommender.content = self.content_mock
        
        hybrid_recs, hybrid_rec_names = self.recommender.make_content_recs(1, m=3)
        
        # Test that the correct methods were called
        mock_create_user_item_matrix.assert_called_once()
        mock_get_user_articles.assert_called_once_with(1, self.interactions_mock)
        mock_get_article_names.assert_called_once()
        
        # Verify the recommendations
        self.assertEqual(hybrid_recs, [105, 103, 104])
        self.assertEqual(hybrid_rec_names, ['Article E', 'Article C', 'Article D'])

    @patch('recommender_ibmws.recommender.rf.get_top_article_ids')
    def test_cold_start_user(self, mock_get_top_article_ids):
        # Simulate cold start by testing with a new user ID (not in interactions)
        mock_get_top_article_ids.return_value = [105, 103, 104]
        
        self.recommender.interactions = self.interactions_mock
        self.recommender.content = self.content_mock
        
        hybrid_recs, hybrid_rec_names = self.recommender.make_content_recs(999, m=3)
        
        # Verify that the fallback method was used
        mock_get_top_article_ids.assert_called_once_with(3, self.interactions_mock)
        self.assertEqual(hybrid_recs, [105, 103, 104])

    @patch('recommender_ibmws.recommender.rf.create_user_item_matrix')
    def test_no_user_interactions(self, mock_create_user_item_matrix):
        # Test behavior when user has no interactions
        mock_create_user_item_matrix.return_value = pd.DataFrame({
            101: [0, 0, 0, 0],
            102: [0, 0, 0, 0],
            103: [0, 0, 0, 0],
            104: [0, 0, 0, 0],
        }, index=[1, 2, 3, 4])
        
        self.recommender.interactions = self.interactions_mock
        self.recommender.content = self.content_mock
        
        hybrid_recs, hybrid_rec_names = self.recommender.make_content_recs(1, m=3)
        
        # Since no interactions, it should fallback to top article recommendations
        self.assertEqual(len(hybrid_recs), 3)

if __name__ == '__main__':
    unittest.main()


import numpy as np
import pandas as pd
import importlib.resources as pkg_resources
from recommender_ibmws import data

def load_data(inter_path, content_path):
    """
	Loads interaction and content data from CSV files, maps email addresses to unique user IDs, 
	and returns the processed interaction and content dataframes.

	Parameters:
		inter_path (str): The filepath to the interactions CSV file.
		content_path (str): The filepath to the content CSV file.

	Returns:
		tuple: A tuple containing the processed interaction and content dataframes.
	"""
    #interactions = pd.read_csv(inter_path)
    #content = pd.read_csv(content_path)
    with pkg_resources.open_text(data, inter_path) as inter_file:
        interactions = pd.read_csv(inter_file)

    with pkg_resources.open_text(data, content_path) as content_file:
        content = pd.read_csv(content_file)
    
    del interactions['Unnamed: 0']
    del content['Unnamed: 0']
    
    email_encoded = email_mapper(interactions)
    del interactions['email']
    interactions['user_id'] = email_encoded
    
    print(interactions.head())
    print(content.head())
    
    return interactions, content

def email_mapper(interactions):
    """
    Maps email addresses to unique user IDs.

    Parameters:
        interactions (pandas DataFrame): Interaction data.

    Returns:
        list: A list of user IDs corresponding to the input email addresses.
    """
    
    coded_dict = dict()
    cter = 1
    email_encoded = []
    
    for val in interactions['email']:
        if val not in coded_dict:
            coded_dict[val] = cter
            cter+=1
        
        email_encoded.append(coded_dict[val])
    return email_encoded

def get_top_articles(n, interactions):
    """
    Returns the top 'n' article titles based on their view numbers.
    Parameters:
        n (int): The number of top articles to return.
        interactions (pandas dataframe): The dataframe containing article view numbers.
    Returns:
        list: A list of the top 'n' article titles.
    """
    article_view_number_ranked = interactions['title'].value_counts().sort_values(ascending=False)
    
    top_articles = article_view_number_ranked.index[:n].tolist()
    top_articles =[str(i) for i in top_articles]
    
    return top_articles # Return the top article titles from df (not df_content)

def get_top_article_ids(n, interactions):
    """
    Returns the top 'n' article IDs based on their view numbers.
    Parameters:
        n (int): The number of top articles to return.
        interactions (pandas dataframe, optional): The dataframe containing article view numbers. Defaults to the global variable 'df'.
    Returns:
        list: A list of the top 'n' article IDs.
    """
    article_view_number_ranked = interactions['title'].value_counts().sort_values(ascending=False)
    
    top_articles = article_view_number_ranked.index.values[:n]
    top_articles =[str(i) for i in top_articles]
    
    return top_articles # Return the top article ids


def create_user_item_matrix(interactions):
    """
    Creates a user-item matrix from a given DataFrame.
    Parameters:
        interactions (pandas DataFrame): A DataFrame containing user and article interactions.
    Returns:
        pandas DataFrame: A user-item matrix where rows represent users, columns represent articles, and cell values indicate interaction presence.
    """
    # Get unique user and article ids
    users = interactions['user_id'].unique()
    articles = interactions['article_id'].unique()
    
    # Create a dictionary to map a article_ids to column indices
    article_to_idx = {article: idx for idx, article in enumerate(articles)}
    
    # Create and initializa a user-article matrix with zeros
    user_item = np.zeros((len(users), len(articles)))
    
    # Fill the matrix with 1s when interactions happen
    for i, row in interactions.iterrows():
        user_idx = np.where(users == row['user_id'])[0][0]
        article_idx = article_to_idx[row['article_id']]
        user_item[user_idx, article_idx] = 1
        
    # Create a DataFrame from the numpy array
    user_item = pd.DataFrame(user_item, index=users, columns=articles)
    
    return user_item # return the user_item matrix 

def get_article_names(article_ids, interactions):
    """
    Retrieves a list of article names corresponding to the given article IDs.
    
    Parameters:
        article_ids (list): A list of article IDs.
        interactions (pandas DataFrame, optional): A DataFrame containing article information. Defaults to the global variable 'df'.
    
    Returns:
        list: A list of article names.
    """
    # Filter the df to include the requested article ids only
    interactions_filtered = interactions[interactions['article_id'].isin(article_ids)]
    
    # Get the titles for the filteresd article_ids
    article_names = interactions_filtered['title'].unique().tolist()
    
    # Sort the article names in descending order based on the index of article ids
    article_names = interactions_filtered['title'].unique().tolist()
    #article_names = interactions_filtered[interactions_filtered['article_id'].isin(article_ids)].sort_values(by='article_id')['title'].unique().tolist()
    
    return article_names # Return the article names associated with list of article ids


def get_user_articles(user_id, interactions):
    """
    Retrieves a list of article IDs and their corresponding names that a given user has interacted with.
    
    Parameters:
        user_id (int): The ID of the user to retrieve articles for.
        interactions (pandas DataFrame): A DataFrame containing user and article interactions.
    
    Returns:
        tuple: A tuple containing a list of article IDs and a list of article names.
    """
    user_item = create_user_item_matrix(interactions)
    # Get article ids that user_id has interacted
    article_ids = user_item.columns[user_item.loc[user_id] == 1].tolist()
    
    # Sort the article_ids based on their order in the user_item DataFrame
    article_ids.sort(key=lambda x: list(user_item.loc[user_id]).index(1))
    
    # Get article titles corresponding to the article ids retrieved
    article_names = get_article_names(article_ids, interactions)
    
    article_ids=[str(i) for i in article_ids]
    
    return article_ids, article_names # return the ids and names


from setuptools import setup, find_packages

setup(name='recommender_ibmws',
      version='0.1.1',
      author= 'Joana Martins dos Santos',
      author_email= 'joanamlmsantos@gmail.com',
      description='User article recommender system for IBM Watson Studio',
      packages=find_packages(),
      include_package_data=True,
      install_requires=[  # Add this line to specify dependencies
        'pandas>=1.0.0',
        'numpy>=1.18.0',
        'scikit-learn>=0.22.0'
        ],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.12',
    url = "https://github.com/joanamdsantos/recommender_ibmws_pkg",
    zip_safe=False)
